/*
Copyright (c) 2011, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <grub/dl.h>
#include <grub/env.h>
#include <grub/err.h>
#include <grub/extcmd.h>
#include <grub/misc.h>

#include "datatype.h"
#include "bitsutil.h"

#include "acpi.h"
#include "accommon.h"
#include "acnamesp.h"
#include "amlresrc.h"

ACPI_MODULE_NAME("grub2-acpica")

static U32 acpica_init_state = 0;
static bool acpica_cpus_initialized = false;
static U32 acpica_cpus_init_caps = 0;

static bool IsEnabledProcessor(ACPI_HANDLE ObjHandle)
{
    bool ret = false;
    ACPI_DEVICE_INFO *Info;

    if (ACPI_SUCCESS(AcpiGetObjectInfo(ObjHandle, &Info)))
        if ((Info->Type == ACPI_TYPE_PROCESSOR) && (Info->Valid & ACPI_VALID_STA) && (Info->CurrentStatus & ACPI_STA_DEVICE_ENABLED))
            ret = true;
    ACPI_FREE(Info);

    return ret;
}

static grub_err_t init(void)
{
    if (acpica_init_state == 1)
        return GRUB_ERR_NONE;

    if (AcpiInitializeTables(NULL, 0, 0) != AE_OK)
        return GRUB_ERR_IO;

    if (AcpiInitializeSubsystem() != AE_OK)
        return GRUB_ERR_IO;

    if (AcpiLoadTables() != AE_OK)
        return GRUB_ERR_IO;

    if (AcpiEnableSubsystem(ACPI_FULL_INITIALIZATION) != AE_OK)
        return GRUB_ERR_IO;

    if (AcpiInitializeObjects(ACPI_FULL_INITIALIZATION) != AE_OK)
        return GRUB_ERR_IO;

    acpica_init_state = 1;

    return GRUB_ERR_NONE;
}

static grub_err_t grub_cmd_acpi_csd(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *csd_obj;
    ACPI_OBJECT *depend_obj;
    U32 i;
    U32 j;
    U32 depend_count;
    U32 dword_count;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._CSD\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_CSD", NULL, &Results, ACPI_TYPE_PACKAGE);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _CSD failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    csd_obj = (ACPI_OBJECT *)Results.Pointer;

    depend_count = csd_obj->Package.Count;

    dprintf("acpica", "dependency count = %u\n", depend_count);

    grub_printf("               %-10s %-10s %-10s %-10s %-10s    %-10s\n",
                "NumEntries", "Revision", "Domain", "CoordType", "NumProcessors", "Index");

    for (i = 0; i < depend_count; i++) {
        depend_obj = &csd_obj->Package.Elements[i];
        if (depend_obj->Type != ACPI_TYPE_PACKAGE) {
            grub_printf("Dependency is not a package\n");
            continue;
        }

        dword_count = depend_obj->Package.Count;
        if (depend_obj->Package.Count != 6) {
            grub_printf("Dependency does not include 6 entries\n");
            continue;
        }

        if (depend_obj->Package.Elements[0].Type != ACPI_TYPE_INTEGER) {
            grub_printf("Dependency entry 0 is not an integer\n");
            continue;
        }

        if (depend_obj->Package.Elements[0].Integer.Value != 6) {
            grub_printf("Dependency number of entries is not 6\n");
            continue;
        }

        if (depend_obj->Package.Elements[1].Type != ACPI_TYPE_INTEGER) {
            grub_printf("Dependency entry 1 is not an integer\n");
            continue;
        }

        if (depend_obj->Package.Elements[1].Integer.Value != 0) {
            grub_printf("Dependency revision is not 0\n");
            continue;
        }

        grub_printf("Dependency[%u]:", i);

        for (j = 0; j < dword_count; j++) {
            if (depend_obj->Package.Elements[j].Type != ACPI_TYPE_INTEGER) {
                grub_printf("Dependency entry %u is not an integer\n", j);
                continue;
            }
            grub_printf(" 0x%08llx", depend_obj->Package.Elements[j].Integer.Value);
        }
        grub_printf("\n");
    }

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static void print_resource_decode(AML_RESOURCE_GENERIC_REGISTER *reg)
{
    grub_printf(" Resource: ");
    if (reg->AddressSpaceId == 0x7f) {
        if ((reg->BitWidth == 0) && (reg->BitOffset == 0) && (reg->AccessSize == 0) && (reg->Address == 0)) {
            grub_printf("FFH C1 Halt\n");
            return;
        }
        if (reg->BitWidth == 1) {
            if (reg->BitOffset == 1)
                return;
            if (reg->BitOffset == 2)
                grub_printf("Native MWAIT; hint 0x%02llx; %sHW coord; %s BM avoidance\n",
                            reg->Address,
                            (reg->AccessSize & (1 << 0)) ? "" : "no ",
                            (reg->AccessSize & (1 << 1)) ? "use" : "no");
        }
    }
}

static grub_err_t grub_cmd_acpi_cst(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *cst_obj;
    ACPI_OBJECT *cstate_obj;
    U32 i, j, cstate_count;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._CST\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_CST", NULL, &Results, ACPI_TYPE_PACKAGE);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _CST failed (0x%x %s)", Status, AcpiFormatException(Status));
        goto out;
    }

    cst_obj = (ACPI_OBJECT *)Results.Pointer;

    cstate_count = cst_obj->Package.Count;
    cstate_obj = &cst_obj->Package.Elements[0];

    dprintf("acpica", "cstate_count=0x%x\n", cstate_count);

    if (cstate_obj->Type != ACPI_TYPE_INTEGER || cstate_obj->Integer.Value != cstate_count - 1) {
        grub_printf("CState count inconsistent with containing package length\n");
        goto out;
    }

    grub_printf("          %-23s %-10s %-10s %-10s\n",
                "Resource", "Type", "Latency", "Power");

    for (i = 1; i < cstate_count; i++) {
        AML_RESOURCE_GENERIC_REGISTER *reg;

        cstate_obj = &cst_obj->Package.Elements[i];
        if (cstate_obj->Type != ACPI_TYPE_PACKAGE) {
            grub_printf("CState was not a package\n");
            continue;
        }
        if (cstate_obj->Package.Count != 4) {
            grub_printf("CState package length != 4\n");
            continue;
        }
        if (cstate_obj->Package.Elements[0].Type != ACPI_TYPE_BUFFER) {
            grub_printf("CState doesn't start with a buffer (type=%u)\n", cstate_obj->Package.Elements[0].Type);
            continue;
        }

        reg = (AML_RESOURCE_GENERIC_REGISTER *)(cstate_obj->Package.Elements[0].Buffer.Pointer);
        grub_printf("CState %u:", i);
        grub_printf(" (0x%x,0x%x,0x%x,0x%x,0x%02llx)", reg->AddressSpaceId, reg->BitWidth, reg->BitOffset, reg->AccessSize, reg->Address);
        for (j = 1; j < 4; j++)
            grub_printf(" 0x%08llx", cstate_obj->Package.Elements[j].Integer.Value);
        grub_printf("\n");
        print_resource_decode(reg);
        grub_printf(" Type: ACPI C%llu\n", cstate_obj->Package.Elements[1].Integer.Value);
        grub_printf(" Latency: %llu microseconds\n", cstate_obj->Package.Elements[2].Integer.Value);
        grub_printf(" Power: %llu mW\n", cstate_obj->Package.Elements[3].Integer.Value);
    }

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_pct(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *pct_obj;
    U32 i;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._PCT\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_PCT", NULL, &Results, ACPI_TYPE_PACKAGE);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _PCT failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    pct_obj = (ACPI_OBJECT *)Results.Pointer;

    if (pct_obj->Package.Count != 2) {
        grub_printf("PCT does not contain 2 resources\n");
        goto out;
    }

    for (i = 0; i < 2; i++) {
        AML_RESOURCE_GENERIC_REGISTER *reg;

        if (pct_obj->Package.Elements[0].Type != ACPI_TYPE_BUFFER) {
            grub_printf("PCT resource doesn't start with a buffer (type=%u)\n", pct_obj->Package.Elements[i].Type);
            continue;
        }

        reg = (AML_RESOURCE_GENERIC_REGISTER *)(pct_obj->Package.Elements[i].Buffer.Pointer);
        grub_printf("Register %u:", i);
        grub_printf(" (0x%x,0x%x,0x%x,0x%x,0x%02llx)\n", reg->AddressSpaceId, reg->BitWidth, reg->BitOffset, reg->AccessSize, reg->Address);
    }

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_processor(struct grub_extcmd *cmd, int argc, char **args)
{
    ACPI_HANDLE cpu_handle;
    ACPI_DEVICE_INFO *Info = NULL;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    if (ACPI_FAILURE(AcpiGetObjectInfo(cpu_handle, &Info)))
        return grub_error(GRUB_ERR_IO, "Could not get info for specified CPU");

    if (Info->Valid & ACPI_VALID_HID)
        grub_printf(" _HID: %s\n", Info->HardwareId.String);

    if (Info->Valid & ACPI_VALID_UID)
        grub_printf(" _UID: %s\n", Info->UniqueId.String);

    if (Info->Valid & ACPI_VALID_ADR)
        grub_printf(" _ADR: 0x%llx\n", Info->Address);

    if (Info->Valid & ACPI_VALID_STA) {
        grub_printf(" _STA: 0x%x (%s, %s, %s)\n",
            Info->CurrentStatus,
            (Info->CurrentStatus & ACPI_STA_DEVICE_PRESENT) ? "Present" : "Not Present",
            (Info->CurrentStatus & ACPI_STA_DEVICE_ENABLED) ? "Enabled" : "Not Enabled",
            (Info->CurrentStatus & ACPI_STA_DEVICE_OK) ? "OK" : "Not OK");
    }

    ACPI_FREE(Info);
    return GRUB_ERR_NONE;
}

static grub_err_t grub_cmd_acpi_pdl(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *pdl_obj;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._PDL\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_PDL", NULL, &Results, ACPI_TYPE_INTEGER);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _PDL failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    pdl_obj = (ACPI_OBJECT *)Results.Pointer;
    grub_printf("Lowest performance state that OSPM can use = %llu\n", pdl_obj->Integer.Value);

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_ppc(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *ppc_obj;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._PPC\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_PPC", NULL, &Results, ACPI_TYPE_INTEGER);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _PPC failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    ppc_obj = (ACPI_OBJECT *)Results.Pointer;
    grub_printf("Highest performance state that OSPM can use = %llu\n", ppc_obj->Integer.Value);

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_psd(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *psd_obj;
    ACPI_OBJECT *depend_obj;
    U32 i;
    U32 j;
    U32 depend_count;
    U32 dword_count;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._PSD\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_PSD", NULL, &Results, ACPI_TYPE_PACKAGE);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _PSD failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    psd_obj = (ACPI_OBJECT *)Results.Pointer;

    depend_count = psd_obj->Package.Count;

    dprintf("acpica", "dependency count = %u\n", depend_count);

    grub_printf("               %-10s %-10s %-10s %-10s %-10s\n",
                "NumEntries", "Revision", "Domain", "CoordType", "NumProcessors");

    for (i = 0; i < depend_count; i++) {
        depend_obj = &psd_obj->Package.Elements[i];
        if (depend_obj->Type != ACPI_TYPE_PACKAGE) {
            grub_printf("Dependency is not a package\n");
            continue;
        }

        dword_count = depend_obj->Package.Count;
        if (depend_obj->Package.Count != 5) {
            grub_printf("Dependency does not include 5 entries\n");
            continue;
        }

        if (depend_obj->Package.Elements[0].Type != ACPI_TYPE_INTEGER) {
            grub_printf("Dependency entry 0 is not an integer\n");
            continue;
        }

        if (depend_obj->Package.Elements[0].Integer.Value != 5) {
            grub_printf("Dependency number of entries is not 5\n");
            continue;
        }

        if (depend_obj->Package.Elements[1].Type != ACPI_TYPE_INTEGER) {
            grub_printf("Dependency entry 1 is not an integer\n");
            continue;
        }

        if (depend_obj->Package.Elements[1].Integer.Value != 0) {
            grub_printf("Dependency revision is not 0\n");
            continue;
        }

        grub_printf("Dependency[%u]:", i);
        for (j = 0; j < dword_count; j++) {
            if (depend_obj->Package.Elements[j].Type != ACPI_TYPE_INTEGER) {
                grub_printf("Dependency entry %u is not an integer\n", j);
                continue;
            }
            grub_printf(" 0x%08llx", depend_obj->Package.Elements[j].Integer.Value);
        }
        grub_printf("\n");
    }

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_pss(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *pss_obj;
    ACPI_OBJECT *pstate_obj;
    U32 i, j, pstate_count, dword_count;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._PSS\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_PSS", NULL, &Results, ACPI_TYPE_PACKAGE);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _PSS failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    pss_obj = (ACPI_OBJECT *)Results.Pointer;

    pstate_count = pss_obj->Package.Count;

    dprintf("acpica", "pstate_count=0x%x\n", pstate_count);

    grub_printf("    %-10s %-10s %-10s %-10s %-10s %-10s\n",
                "Frequency", "Power", "Latency", "BMLatency", "Control", "Status");
    for (i = 0; i < pstate_count; i++) {
        pstate_obj = &pss_obj->Package.Elements[i];
        if (pstate_obj->Type != ACPI_TYPE_PACKAGE) {
            grub_printf("PState was not a package\n");
            continue;
        }
        if (pstate_obj->Package.Count != 6) {
            grub_printf("PState package length != 6\n");
            continue;
        }

        dword_count = pstate_obj->Package.Count;
        grub_printf("P%-2u", i);

        for (j = 0; j < dword_count; j++)
            grub_printf(" 0x%08llx", pstate_obj->Package.Elements[j].Integer.Value);
        grub_printf("\n");
    }

out:
    ACPI_FREE(Results.Pointer);

    return ret;
}

static grub_err_t grub_cmd_acpi_ptc(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *ptc_obj;
    U32 i;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._PTC\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_PTC", NULL, &Results, ACPI_TYPE_PACKAGE);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _PTC failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    ptc_obj = (ACPI_OBJECT *)Results.Pointer;

    if (ptc_obj->Package.Count != 2) {
        grub_printf("PTC does not contain 2 resources\n");
        goto out;
    }

    for (i = 0; i < 2; i++) {
        AML_RESOURCE_GENERIC_REGISTER *reg;

        if (ptc_obj->Package.Elements[0].Type != ACPI_TYPE_BUFFER) {
            grub_printf("PTC resource doesn't start with a buffer (type=%u)\n", ptc_obj->Package.Elements[i].Type);
            continue;
        }

        reg = (AML_RESOURCE_GENERIC_REGISTER *)(ptc_obj->Package.Elements[i].Buffer.Pointer);
        grub_printf("Register %u:", i);
        grub_printf(" (0x%x,0x%x,0x%x,0x%x,0x%02llx)\n", reg->AddressSpaceId, reg->BitWidth, reg->BitOffset, reg->AccessSize, reg->Address);
    }

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_tpc(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *tpc_obj;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._TPC\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_TPC", NULL, &Results, ACPI_TYPE_INTEGER);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _TPC failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    tpc_obj = (ACPI_OBJECT *)Results.Pointer;
    grub_printf("Highest throttling state that OSPM can use = %llu\n", tpc_obj->Integer.Value);

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_tdl(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *tdl_obj;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._TDL\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_TDL", NULL, &Results, ACPI_TYPE_INTEGER);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _TDL failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    tdl_obj = (ACPI_OBJECT *)Results.Pointer;
    grub_printf("Lowest power throttling state that OSPM may use = %llu\n", tdl_obj->Integer.Value);

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_tsd(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *tsd_obj;
    ACPI_OBJECT *depend_obj;
    U32 i;
    U32 j;
    U32 depend_count;
    U32 dword_count;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._TSD\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_TSD", NULL, &Results, ACPI_TYPE_PACKAGE);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _TSD failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    tsd_obj = (ACPI_OBJECT *)Results.Pointer;

    depend_count = tsd_obj->Package.Count;

    dprintf("acpica", "dependency count = %u\n", depend_count);

    grub_printf("               %-10s %-10s %-10s %-10s %-10s\n",
                "NumEntries", "Revision", "Domain", "CoordType", "NumProcessors");

    for (i = 0; i < depend_count; i++) {
        depend_obj = &tsd_obj->Package.Elements[i];
        if (depend_obj->Type != ACPI_TYPE_PACKAGE) {
            grub_printf("Dependency is not a package\n");
            continue;
        }

        dword_count = depend_obj->Package.Count;
        if (depend_obj->Package.Count != 5) {
            grub_printf("Dependency does not include 5 entries\n");
            continue;
        }

        if (depend_obj->Package.Elements[0].Type != ACPI_TYPE_INTEGER) {
            grub_printf("Dependency entry 0 is not an integer\n");
            continue;
        }

        if (depend_obj->Package.Elements[0].Integer.Value != 5) {
            grub_printf("Dependency number of entries is not 5\n");
            continue;
        }

        if (depend_obj->Package.Elements[1].Type != ACPI_TYPE_INTEGER) {
            grub_printf("Dependency entry 1 is not an integer\n");
            continue;
        }

        if (depend_obj->Package.Elements[1].Integer.Value != 0) {
            grub_printf("Dependency revision is not 0\n");
            continue;
        }

        grub_printf("Dependency[%u]:", i);
        for (j = 0; j < dword_count; j++) {
            if (depend_obj->Package.Elements[j].Type != ACPI_TYPE_INTEGER) {
                grub_printf("Dependency entry %u is not an integer\n", j);
                continue;
            }
            grub_printf(" 0x%08llx", depend_obj->Package.Elements[j].Integer.Value);
        }
        grub_printf("\n");
    }

out:
    ACPI_FREE(Results.Pointer);
    return ret;
}

static grub_err_t grub_cmd_acpi_tss(struct grub_extcmd *cmd, int argc, char **args)
{
    grub_err_t ret;
    ACPI_HANDLE cpu_handle;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT *tss_obj;
    ACPI_OBJECT *tstate_obj;
    U32 i, j, tstate_count, dword_count;

    (void)cmd;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: \\PATH._TO_.CPU_");

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    grub_printf("\n%s._TSS\n", args[0]);

    if (ACPI_FAILURE(AcpiGetHandle(NULL, args[0], &cpu_handle)))
        return grub_error(GRUB_ERR_IO, "Could not get handle to specified CPU");

    Status = AcpiEvaluateObjectTyped(cpu_handle, "_TSS", NULL, &Results, ACPI_TYPE_PACKAGE);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND || Status == AE_NULL_OBJECT) {
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
        grub_printf(" Not found\n");
        goto out;
    } else {
        ret = grub_error(GRUB_ERR_IO, "Evaluating _TSS failed (0x%x %s)\n", Status, AcpiFormatException(Status));
        goto out;
    }

    tss_obj = (ACPI_OBJECT *)Results.Pointer;

    tstate_count = tss_obj->Package.Count;

    dprintf("acpica", "tstate_count=0x%x\n", tstate_count);

    grub_printf("    %-10s %-10s %-10s %-10s %-10s\n",
                "Percent", "Power", "Latency", "Control", "Status");
    for (i = 0; i < tstate_count; i++) {
        tstate_obj = &tss_obj->Package.Elements[i];
        if (tstate_obj->Type != ACPI_TYPE_PACKAGE) {
            grub_printf("TState was not a package\n");
            continue;
        }
        if (tstate_obj->Package.Count != 5) {
            grub_printf("TState package length != 5\n");
            continue;
        }

        dword_count = tstate_obj->Package.Count;
        grub_printf("T%-2u", i);

        for (j = 0; j < dword_count; j++)
            grub_printf(" 0x%08llx", tstate_obj->Package.Elements[j].Integer.Value);
        grub_printf("\n");
    }

out:
    ACPI_FREE(Results.Pointer);

    return ret;
}

static grub_err_t call_osc(ACPI_HANDLE cpu_handle, U32 caps)
{
    grub_err_t ret;
    ACPI_STATUS Status;
    ACPI_BUFFER Results = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };
    ACPI_OBJECT_LIST Params;
    ACPI_OBJECT Obj[4];
    U32 osc_buffer[2];
    U8 OSC_UUID[16] = {0x16, 0xA6, 0x77, 0x40, 0x0C, 0x29, 0xBE, 0x47, 0x9E, 0xBD, 0xD8, 0x70, 0x58, 0x71, 0x39, 0x53};

    /* Initialize the parameter list */
    Params.Count = 4;
    Params.Pointer = Obj;

    /* Initialize the parameter objects */
    // Intel-specific UUID
    Obj[0].Type = ACPI_TYPE_BUFFER;
    Obj[0].Buffer.Length = 16;
    Obj[0].Buffer.Pointer = OSC_UUID;

    // Revision ID
    Obj[1].Type = ACPI_TYPE_INTEGER;
    Obj[1].Integer.Value = 1;

    // Count of DWORDS
    Obj[2].Type = ACPI_TYPE_INTEGER;
    Obj[2].Integer.Value = 2;

    // Capabilities buffer
    Obj[3].Type = ACPI_TYPE_BUFFER;
    Obj[3].Buffer.Length = 8;
    Obj[3].Buffer.Pointer = (U8 *)osc_buffer;

    osc_buffer[0] = 0;
    osc_buffer[1] = caps; // Capabilities DWORD

    Status = AcpiEvaluateObject(cpu_handle, "_OSC", &Params, &Results);
    if (Status == AE_OK)
        ret = GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND)
        ret = grub_error(GRUB_ERR_TEST_FAILURE, "false");
    else
        ret = grub_error(GRUB_ERR_IO, "Evaluating _OSC failed (0x%x %s)\n", Status, AcpiFormatException(Status));

    ACPI_FREE(Results.Pointer);

    return ret;
}

static grub_err_t call_pdc(ACPI_HANDLE cpu_handle, U32 caps)
{
    ACPI_STATUS Status;
    ACPI_OBJECT_LIST Params;
    ACPI_OBJECT Obj;
    U32 pdc_buffer[3];

    /* Initialize the parameter list */
    Params.Count = 1;
    Params.Pointer = &Obj;

    /* Initialize the parameter objects */
    Obj.Type = ACPI_TYPE_BUFFER;
    Obj.Buffer.Length = 12;
    Obj.Buffer.Pointer = (UINT8 *)pdc_buffer;

    /* Initialize the DWORD array */
    pdc_buffer[0] = 1; // Revision
    pdc_buffer[1] = 1; // Count
    pdc_buffer[2] = caps; // Capabilities DWORD

    Status = AcpiEvaluateObject(cpu_handle, "_PDC", &Params, NULL);
    if (Status == AE_OK)
        return GRUB_ERR_NONE;
    else if (Status == AE_NOT_FOUND)
        return grub_error(GRUB_ERR_TEST_FAILURE, "false");
    else
        return grub_error(GRUB_ERR_IO, "Evaluating _PDC failed (0x%x %s)\n", Status, AcpiFormatException(Status));
}

struct find_processor_context {
    bool init_cpu;
    U32 caps;
    struct buffer *buffer;
};

static ACPI_STATUS find_processor(ACPI_HANDLE ObjHandle, UINT32 NestingLevel ACPI_UNUSED_VAR, void *Context, void **ReturnValue ACPI_UNUSED_VAR)
{
    struct find_processor_context *fpc = Context;
    struct buffer *buffer = fpc->buffer;
    ACPI_BUFFER Path = { .Length = ACPI_ALLOCATE_BUFFER, .Pointer = NULL };

    if (!IsEnabledProcessor(ObjHandle))
        goto out;

    if (ACPI_FAILURE(AcpiGetName(ObjHandle, ACPI_FULL_PATHNAME, &Path))) {
        grub_printf("Couldn't get object name\n");
        goto out;
    }

    if (buffer->len != 0)
        buffer_append_str(buffer, " ");
    buffer_append_str(buffer, Path.Pointer);

    if (fpc->init_cpu)
        if (call_osc(ObjHandle, fpc->caps) != GRUB_ERR_NONE)
            call_pdc(ObjHandle, fpc->caps);

out:
    ACPI_FREE(Path.Pointer);
    return AE_OK;
}

static const struct grub_arg_option acpi_find_cpus_options[] = {
#undef OPTION_VARNAME
#define OPTION_VARNAME 0
    {"varname", 'v', 0, "Save read value into variable VARNAME", "VARNAME", ARG_TYPE_STRING},
#undef OPTION_CAPS
#define OPTION_CAPS 1
    {"caps", 'c', 0, "Capabilities DWORD (default=0xfbf)", "DWORD", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_acpi_find_cpus(struct grub_extcmd *cmd, int argc, char **args)
{
    struct buffer buffer = BUFFER_INIT;
    struct find_processor_context fpc = { .init_cpu = false, .caps = 0xfbf, .buffer = &buffer };

    (void)argc;
    (void)args;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CAPS].set)
        if (strtou32_h(cmd->state[OPTION_CAPS].arg, &fpc.caps) != GRUB_ERR_NONE)
            return grub_errno;

    if (acpica_cpus_initialized) {
        if (fpc.caps != acpica_cpus_init_caps)
            return grub_error(GRUB_ERR_IO, "CPUs already initialized with caps 0x%x; cannot reinit with 0x%x; call acpi_terminate first", acpica_cpus_init_caps, fpc.caps);
    } else {
        acpica_cpus_initialized = true;
        acpica_cpus_init_caps = fpc.caps;
        fpc.init_cpu = true;
    }

    AcpiWalkNamespace(ACPI_TYPE_PROCESSOR, ACPI_ROOT_OBJECT, ACPI_UINT32_MAX, find_processor, NULL, &fpc, NULL);

    buffer_append_nul(&buffer);

    if (cmd->state[OPTION_VARNAME].set)
        grub_env_set(cmd->state[OPTION_VARNAME].arg, buffer.ptr);
    else
        grub_printf("%s\n", (char *)buffer.ptr);

    buffer_free(&buffer);

    return GRUB_ERR_NONE;
}

static grub_err_t grub_cmd_acpi_terminate(struct grub_extcmd *cmd, int argc, char **args)
{
    (void)cmd;
    (void)argc;
    (void)args;

    AcpiTerminate();
    acpica_init_state = 0;
    acpica_cpus_initialized = false;

    return GRUB_ERR_NONE;
}

static grub_extcmd_t cmd_acpi_csd;
static grub_extcmd_t cmd_acpi_cst;
static grub_extcmd_t cmd_acpi_pct;
static grub_extcmd_t cmd_acpi_processor;
static grub_extcmd_t cmd_acpi_pdl;
static grub_extcmd_t cmd_acpi_ppc;
static grub_extcmd_t cmd_acpi_psd;
static grub_extcmd_t cmd_acpi_pss;
static grub_extcmd_t cmd_acpi_ptc;
static grub_extcmd_t cmd_acpi_tdl;
static grub_extcmd_t cmd_acpi_tpc;
static grub_extcmd_t cmd_acpi_tsd;
static grub_extcmd_t cmd_acpi_tss;
static grub_extcmd_t cmd_acpi_find_cpus;
static grub_extcmd_t cmd_acpi_terminate;

GRUB_MOD_INIT(acpica)
{
    // Bit field that enables/disables debug output from entire subcomponents within the ACPICA subsystem.
    // AcpiDbgLevel = 0;

    // Bit field that enables/disables the various debug output levels
    // AcpiDbgLayer = 0;

    dprintf("acpica", "ACPI_CA_VERSION = %x\n", ACPI_CA_VERSION);

    cmd_acpi_csd = grub_register_extcmd("acpi_csd", grub_cmd_acpi_csd, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _CSD (C-State Dependency) method on a CPU and decode the result",
                                        0);
    cmd_acpi_cst = grub_register_extcmd("acpi_cst", grub_cmd_acpi_cst, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _CST (C-States) method on a CPU and decode the result",
                                        0);
    cmd_acpi_pct = grub_register_extcmd("acpi_pct", grub_cmd_acpi_pct, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _PCT (Performance Control) method on a CPU and decode the result",
                                        0);
    cmd_acpi_processor = grub_register_extcmd("acpi_processor", grub_cmd_acpi_processor, GRUB_COMMAND_FLAG_BOTH,
                                              "\\PATH._TO_.CPU_",
                                              "Display information about a CPU object",
                                              0);
    cmd_acpi_pdl = grub_register_extcmd("acpi_pdl", grub_cmd_acpi_pdl, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _PDL (P-State Depth Limit) method on a CPU and decode the result",
                                        0);
    cmd_acpi_ppc = grub_register_extcmd("acpi_ppc", grub_cmd_acpi_ppc, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _PPC (Performance Present Capabilities) method on a CPU and decode the result",
                                        0);
    cmd_acpi_psd = grub_register_extcmd("acpi_psd", grub_cmd_acpi_psd, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _PSD (P-State Dependency) method on a CPU and decode the result",
                                        0);
    cmd_acpi_pss = grub_register_extcmd("acpi_pss", grub_cmd_acpi_pss, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _PSS (Performance Supported States) method on a CPU and decode the result",
                                        0);
    cmd_acpi_ptc = grub_register_extcmd("acpi_ptc", grub_cmd_acpi_ptc, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _PTC (Processor Throttling Control) method on a CPU and decode the result",
                                        0);
    cmd_acpi_tdl = grub_register_extcmd("acpi_tdl", grub_cmd_acpi_tdl, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _TDL (T-State Depth Limit) method on a CPU and decode the result",
                                        0);
    cmd_acpi_tpc = grub_register_extcmd("acpi_tpc", grub_cmd_acpi_tpc, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _TPC (Throttling Present Capabilities) method on a CPU and decode the result",
                                        0);
    cmd_acpi_tsd = grub_register_extcmd("acpi_tsd", grub_cmd_acpi_tsd, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _TSD (T-State Dependency) method on a CPU and decode the result",
                                        0);
    cmd_acpi_tss = grub_register_extcmd("acpi_tss", grub_cmd_acpi_tss, GRUB_COMMAND_FLAG_BOTH,
                                        "\\PATH._TO_.CPU_",
                                        "Call _TSS (Throttling Supported States) method on a CPU and decode the result",
                                        0);
    cmd_acpi_find_cpus = grub_register_extcmd("acpi_find_cpus", grub_cmd_acpi_find_cpus, GRUB_COMMAND_FLAG_BOTH,
                                              "[-c CAPS] [-v VARNAME]",
                                              "Find ACPI CPUs and initialize them via _OSC or _PDC",
                                              acpi_find_cpus_options);
    cmd_acpi_terminate = grub_register_extcmd("acpi_terminate", grub_cmd_acpi_terminate, GRUB_COMMAND_FLAG_BOTH,
                                              "",
                                              "Perform ACPICA module terminate",
                                              0);
}

GRUB_MOD_FINI(acpica)
{
    AcpiTerminate();
    grub_unregister_extcmd(cmd_acpi_csd);
    grub_unregister_extcmd(cmd_acpi_cst);
    grub_unregister_extcmd(cmd_acpi_pct);
    grub_unregister_extcmd(cmd_acpi_processor);
    grub_unregister_extcmd(cmd_acpi_pdl);
    grub_unregister_extcmd(cmd_acpi_ppc);
    grub_unregister_extcmd(cmd_acpi_psd);
    grub_unregister_extcmd(cmd_acpi_pss);
    grub_unregister_extcmd(cmd_acpi_ptc);
    grub_unregister_extcmd(cmd_acpi_tdl);
    grub_unregister_extcmd(cmd_acpi_tsd);
    grub_unregister_extcmd(cmd_acpi_tss);
    grub_unregister_extcmd(cmd_acpi_find_cpus);
    grub_unregister_extcmd(cmd_acpi_terminate);
}
